#ifndef h_GlStuff
#define h_GlStuff

// OpenGL
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>

#include "ParameterAPI.h"
#include "Parameters.h"

#include <cmath>
#include <deque>
#include <iostream>

#define __PI    3.1415926

#define ABS(x)  (float) (x < 0 ? -x : x)
#define COS(x)  (float) cos( (double) (x) * __PI / 180.0 )
#define SIN(x)  (float) sin( (double) (x) * __PI / 180.0 )


class CloudEvent;

namespace GLStuff
{

// gl callbacks
void GLStartup(int argc, const char** argv);
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void motionFunc(int x, int y);
void timerCallback(int v);
void initialize_graphics( );
void createMenu(void);
void menu(int value);

void DrawString(int x,int y, std::string& s, void* font, float scaleFactor = 0.2f);
void createcircle (int k, int r, int h);
void generateBalls();
GLfloat rand2f( float a, float b );
void addBall();

void go2d();
void go3d();

//-----------------------------------------------------------------------------
// global variables and #defines
//-----------------------------------------------------------------------------

static int numBalls;//sets the number of balls
static int g_speed; //sets the speed of the balls
static int g_bounces; //the number of bounces before the gravitied balls disappear
static float gravity; //the degree of gravity (must be negative)
static float friction;
static float spring; //just leave these as they are
static float g_diameter; //ball diameter
static int g_combineBeforeDrop; //number of times combinations take place before the balls drop
static float g_X ;
static float g_Y ;
static float g_ZOOM; //initial states for the gl_view


// width and height of the window
static GLsizei gWidth;
static GLsizei gHeight;
static GLsizei gLastWidth;
static GLsizei gLastHeight;

// light 0 position
static GLfloat g_light0_pos[4];

//struct for creating the circle shape
typedef struct
{
   float x;
   float y;
}CIRCLE;

static CIRCLE circle;

//struct for holding x and y position of a grain
typedef struct
{
   float xpos;
   float ypos;
}GRAIN_POSITION;

static GRAIN_POSITION*grain_position = new GRAIN_POSITION[numBalls];



class Ball {
public:
	Ball();
	Ball(float xin, float yin, float din, int idin, Ball* oin, bool gravityin, int bouncesin);
	
   ~Ball();
   
	void collide();
	void combine();
	void move();
	void display();
	
	float xpos,ypos,diameter,vxpos,vypos;
	int id, bounces;
	bool gravityON;
	Ball* others;
   
   CloudEvent* fCloudEvent;
   
   int yposAtTimeOfCombination;
   
};

   static Ball*balls;

   static GLboolean gDisplay;
   static GLboolean gFullscreen;

   static int gBufferSize;

   static float gFrameRate;
   static float gRefreshRate;

   static int gMenuItem;
   static GLfloat gInc_val_mouse;

   static bool gLeftButtondown;
   static int gMouseOriginX;
   static int gMouseOriginY;

// experimental
   static int gHopCount;
   static bool gRings;
   static int gHop;

   static bool gShowHUD;

class GLDisplay
{
public:
   GLDisplay(float x, float y, float w, float h)
   : fX(x)
   , fY(y)
   , fW(w)
   , fH(h)
   , fZ(0)
   , fR(1.0)
   , fG(0)
   , fB(0)
   , fA(1.1)
   , fVisible(true)
   , fInitialZAngle(0)
   {
   }
   
   void Display()
   {
      glPushMatrix();
      glTranslatef(fX, fY, fZ);
      glScalef(fW,fH, 1);
      glColor4f(fR, fG, fB, fA);
      Draw();
      glPopMatrix();
   }
   
   virtual void Draw() = 0;
   
   void SetY(float y)
   {
      fY=y;
   }
   
   void SetH(float h)
   {
      fH = h;
   }
   
   void SetZ(float z)
   {
      fZ = z;
   }
   
   void SetInitialZAngle(float phi)
   {
      fInitialZAngle = phi;
   }
   
   void SetColor(float r, float g, float b, float a)
   {
      fR = r; fG = g; fB = b; fA = a;
   }
   
   void SetVisible(bool visible)
   {
      fVisible = visible;
   }
   
   bool Visible() const { return fVisible; }
   
protected:
   float fX, fY, fW, fH, fZ, fR, fG, fB, fA;
   bool fVisible;
   float fInitialZAngle;
};

class Widget
{
public:
   Widget(int x, int y, int w, int h, int parameterId = -1)
   : fX(x)
   , fY(y)
   , fW(w)
   , fH(h)
   , fParameterId(parameterId)
   {
   }
   
   bool Hit(int x, int y)
   {
      
      if ((x >= fX && x <= fX + fW) && (gHeight - y >= fY && gHeight - y <= fY + fH))
      {  
         return true;
      }
      else
      {
         return false;
      }
      
   }
   
   void SetParameterId(int id)
   {
      fParameterId = id;
   }
   
   int X() const { return fX; }
   int Y() const { return fY; }
   int W() const { return fW; }
   int H() const { return fH; }
   
protected:
   int fX, fY, fW, fH;
   int fParameterId;
};

class Slider : public Widget
{
public:
   Slider(int x, int y, int w, int h, int parameterId)
   : Widget(x,y,w,h,parameterId)
   {
      fTouched = false;
      fVal = 0.f;
   }
   
   void Draw()
   {
      go2d();
      
      glPushMatrix();
      glLineWidth(2);

      glColor4f(1,0,0,1);
      glRecti(fX, fY, fX + fW, fY + fH);
      

      
      glColor4f(0,0,0,1);
      glRecti(fX+2, fY+2, fX + fW - 2, fY + fH - 2);

      glColor4f(1,1,1,1);
      glBegin(GL_LINE_STRIP);
      glVertex2f(fX + fVal * fW,  fY + 2);
      glVertex2f(fX + fVal * fW, fY + fH - 2);
      glEnd();
      

      glPopMatrix();
      
      //glScalef(1, fW/fH,1);
      
      //glPushMatrix();
      //glTranslatef(gMouseOriginX/(float)gWidth, 0, 0);
      //glutWireSphere(fH, 5, 3);
      //glPopMatrix();
      
      go3d();
   }
   
   
   
   bool Touched() const { return fTouched; }
   
   void Touch(int x, int y)
   {
      fTouched = true;
      
      int localX = std::max(x - fX, 0);
      
      fVal = std::min(localX, fW) / (float)fW;
      
      if (fParameterId != -1)
         ParameterManager::GetInstance()->GetParameter(fParameterId)->SetValue(fVal);
   }
   
   void UnTouch()
   {
      fTouched = false;
   }
   
   void SetValue(float val)
   {
      fVal = val;
      if (fVal < 0) val = 0;
      if (fVal > 1) val = 1;
   }
   
   float Value() const { return fVal; }
   
private:
   float fVal;
   bool fTouched;
};


class ToggleButton : public Widget
{
public:
   ToggleButton(int x, int y, int w, int h)
   : Widget(x,y,w,h)
   , fToggled(false)
   {
   }
   
   void Draw()
   {
      go2d();
      
      if (fToggled)
      {
         glPushMatrix();
         glColor4f(1,1,1,1);
         glBegin(GL_LINES);
         glVertex2f(fX+2,  fY+2);
         glVertex2f(fX + fW - 2,  fY + fH - 2);
         glVertex2f(fX+2,  fY + fH-2);
         glVertex2f(fX + fW - 2,  fY + 2);
         glEnd();
         glPopMatrix();
      }
      
      glColor4f(0,0,0,1);
      glRecti(fX+2, fY+2, fX + fW - 2, fY + fH - 2);
      glColor4f(1,0,0,1);
      glRecti(fX, fY, fX + fW, fY + fH);
      
      go3d();
   }
   
   void Toggle()
   {
      fToggled = !fToggled;

      if (fParameterId != -1)
         ParameterManager::GetInstance()->GetParameter(fParameterId)->SetValue(fToggled);
   }
   
   bool Toggled() const { return fToggled; }
   
private:
   bool fToggled;
};



}

#endif

