#ifndef h_Insaniac
#define h_Insaniac

#include <iostream>
#include <pthread.h>
#include <vector>

#include "AudioServer.h"
#include "SignalGenerators.h"
#include "WindowFunction.h"
#include "ParameterAPI.h"
#include "Parameters.h"

#include "GLStuff.h"

const int numNotes = 6;

//spooky
const int chord1[numNotes] = { 1, 7, 14, 15, 21, 28};
const int chord2[numNotes] = { 1, 7, 14, 16, 22, 29};

//1 5 8 M3 5 M3
//const int chord1[numNotes] = { 1, 8, 13, 17, 20, 29};
//const int chord2[numNotes] = { 1, 8, 13, 16, 20, 28};

//1 5 9 m3 m7 4
//const int chord1[numNotes] = { 1, 8, 15, 16, 23, 30};
//const int chord2[numNotes] = { 1, 8, 15, 17, 24, 31};

class Grain : public AudioClient
{
public:
   Grain()
   : fWindow(new WindowFunction(4096))
   , fDone(true)
   , fSamplesRendered(0)
   , fTotalSamples(4096)
   , fShouldDeleteMe(false)
   {
      fWindow->setWindowType(WindowFunction::kHann);
      fWindow->setWindowLength(44100);
   }
   
   ~Grain()
   {
      delete fWindow;
   }
   
   void SetLength(float length) // in seconds
   {
      int samps = length * AudioServer::GetInstance()->Fs();
      fWindow->setWindowLength(samps);
   }
   
   void SetDelay(float delay) // in seconds
   {
      int samps = delay * AudioServer::GetInstance()->Fs();
      fTotalSamples = fWindow->getWindowLength() + samps;
      //std::cout << fTotalSamples << std::endl;
   }
   
   void SetFreq(float freq)
   {
      fSinOsc.SetFreq(freq);
   }
   
   void SetWindowType(int type)
   {
      fWindow->setWindowType(type);
   }
   
   bool Done() const { return fDone; }
   void Restart()
   {
      fDone = false;
      fSamplesRendered = 0;
   }
   
   bool ShouldDelete() const { return fShouldDeleteMe; }
   void FlagForDeletion() { fShouldDeleteMe = true; }
   
   virtual void Render(float* buffer, int frames)
   {
      if (!fDone)
      {
         //if (!GLStuff::gToggle.Toggled())
            fSinOsc.Render(buffer, frames);
         //else
         //   fInput.Render(buffer, frames);
         for (int i = 0; i < frames; ++i)
         {
            float f = 0.f;
            if (fSamplesRendered < fWindow->getWindowLength() && !fDone && !fWindow->getNextSample(f)) {
               buffer[i] *= f;
               ++fSamplesRendered;
            }
            else
            {
               ++fSamplesRendered;
               
               if (fSamplesRendered >= fTotalSamples) {
                  fDone = true;
               }
               
               buffer[i] = 0.f;
            }
         }
      }
      
   }
   
private:
   WindowFunction* fWindow;
   SinOsc fSinOsc;
   InputSource fInput;
   bool fDone;
   int fSamplesRendered;
   int fTotalSamples; // grain size + density delay
   bool fShouldDeleteMe;
   
};

class Cloud : public AudioClient
{
public:
   Cloud(float numGrains = 50,
         float minFreq = 2, 
         float maxFreq = 8, 
         float minLength = 0.01,
         float maxLength = 20.f, 
         float density = 0.5)
   : fMaxFreq(maxFreq)
   , fMinFreq(minFreq)
   , fMinLength(minLength)
   , fMaxLength(maxLength)
   , fFreqRange(std::max(fMaxFreq - fMinFreq, 1.f))
   , fDensity(density)
   , fNumGrains(numGrains)
   , fCounter(0)
   , fIsTheDrone(false)  //hack !!! blah!
   {
      for (int i = 0; i < fNumGrains; ++i)
      {
         Grain* g = new Grain;
         g->SetFreq(fMinFreq + rand() % (int)fFreqRange);
         g->SetLength(0.1 + (rand() % 1000) / 1000.f);
         g->SetDelay(.5);
         fGrains.push_back(g);
      }
   }
   
   ~Cloud()
   {
      std::vector<Grain*>::iterator i;
      for (i = fGrains.begin(); i != fGrains.end(); ++i)
      {
         delete (*i);
      }
      
      fGrains.clear();
      
      std::cout << "deleted all grains\n";
   }
   
   void SetNumGrains(int n)
   {
      if (n > fGrains.size())
      {
         for (int i = 0; i < n - fGrains.size(); ++i)
         {
            Grain* g = new Grain;
            g->SetFreq(fMinFreq + rand() % (int)fFreqRange);
            g->SetLength(0.1 + (rand() % 1000) / 1000.f);
            g->SetDelay(.5);
            fGrains.push_back(g);
         }
      }
      else
      {
         int excess = fGrains.size() - n;
         for (int i = 0; i < excess; ++i)
         {
            fGrains.at(i)->FlagForDeletion();
         }
      }
   }
   
   void SetMinFreq(float minFreq)
   {
      fMinFreq = minFreq;
   }
   
   void SetMaxFreq(float maxFreq)
   {
      fMaxFreq = maxFreq;
      fFreqRange = fMaxFreq - fMinFreq;
   }
   
   void SetMinLength(float minLength)
   {
      fMinLength = minLength;
      if (fMaxLength <= fMinLength) fMaxLength = fMinLength;
   }
   
   void SetMaxLength(float maxLength)
   {
      fMaxLength = maxLength;
      if (fMaxLength <= fMinLength) fMaxLength = fMinLength;
   }
   
   void SetDensity(float density)
   {
      fDensity = density;
   }
   
   
   virtual void Render(float* buffer, int frames)
   {      
      float tmp[frames];
      memset(tmp, 0.f, frames*sizeof(float));
      std::vector<Grain*>::iterator i;
      for (i = fGrains.begin(); i != fGrains.end(); ++i)
      {
         (*i)->Render(tmp, frames); // render grain
         
         if ((*i)->Done())
         {
            if ((*i)->ShouldDelete())
            {
               delete *i;
               i = fGrains.erase(i);
            }
            else
            {         
               const float c0 = 8.1757989156;
               float f = 0;
               int r = (rand() % (numNotes-1));
               int baseOctave = rand() % (int)(fMaxFreq) + (int)(fMinFreq);
               int baseOctaveRoot = baseOctave*12;
               f = powf(2, (chord1[r] + baseOctaveRoot) /12.f)*c0;
               if (fCounter++ > 100)
                  f = powf(2, (chord2[r] + baseOctaveRoot) /12.f)*c0;
               
               if (fCounter > 200)
                  fCounter = 0;
               
               //f = f + (rand() / (float)RAND_MAX) * 50;
               
               //(*i)->SetFreq(fMinFreq + rand() % (int)fFreqRange);
               (*i)->SetFreq(f);
               (*i)->SetLength(fMinLength + (rand() / (float)RAND_MAX) * fMaxLength);
               (*i)->SetDelay((rand() / (float)RAND_MAX) * (1-fDensity));
               (*i)->Restart();
               if (this->IsTheDrone())
                  GLStuff::addBall();
            }
         }
         
         // add grain to output
         for (int j = 0; j < frames; ++j)
         {
            buffer[j] += tmp[j];
         }
      }
   }
   
   // HACK!!! DELETE ME!!!! arrrgggh
   void SetIsTheDrone(bool is)
   {
      fIsTheDrone = is;
   }
   
private:
   std::vector<Grain*> fGrains;
   
   float fNumGrains;
   float fMaxFreq;
   float fMinFreq;
   float fMinLength;
   float fMaxLength;
   float fFreqRange;
   float fDensity;
   
   //!!! HACK!!!!! AARRRHJ
   bool fIsTheDrone;
   const bool IsTheDrone() const 
   {
      return fIsTheDrone;
   }
   
   int fCounter;
};

class CloudEvent : public AudioClient
{
public:
   CloudEvent()
   : fGain(1.f)
   , fDone(false)
   , fCloud(Cloud(1,  rand() % 3 + 5, rand() % 4 + 5, .001, .01, (rand() / (float)RAND_MAX*2) + 0.5))
   {
   }
   
   virtual void Render(float* buffer, int frames)
   {
      float tmp[frames];
      memset(tmp, 0.f, frames * sizeof(float));
      fCloud.Render(tmp, frames);
      for (int i = 0; i < frames; ++i)
      {
         buffer[i] = tmp[i] * fGain;
      }
   }
   
   void SetGain(float g)
   {
      fGain = g;
      
      if (fGain > 1.f) fGain = 1.f;
      if (fGain < 0.f) fGain = 0.f;
   }
   
   bool Done() const { return fDone; }
   void SetDone(bool done) { fDone = done; }
   
private:
   Cloud fCloud;
   float fGain;
   bool fDone;
};

class NoiseBlip : public AudioClient
{
public:
   NoiseBlip()
   : fDuration(0.05)
   , fSamplesRendered(0)
   , fDone(false)
   {
   }
   
   virtual void Render(float * buffer, int frames)
   { 
      int durationInSamples = fDuration * AudioServer::GetInstance()->Fs();

      fNoise.Process(buffer, frames);

      for (int i = 0; i < frames; ++i)
      {
         if (fDone)
            buffer[i] = 0;
         else 
         {
            float gain = powf(fSamplesRendered / (float)durationInSamples, 4);
            
            buffer[i] *= gain * 0.5;
               
            ++fSamplesRendered;
            
            if (fSamplesRendered > durationInSamples)
            {
               fDone = true;
               buffer[i] = 0;
            }
         }
      }
      
   }
   
   bool Done() const { return fDone; }
   
private:
   NoiseSource fNoise;

   float fDuration; // seconds
   int fSamplesRendered;
   bool fDone;
};

class Insaniac : public AudioClient
               , public ParameterListener
{
public:
   Insaniac()
   {
      ParameterManager::GetInstance()->GetParameter(kSizeMinParamId)->AddListener(this);
      ParameterManager::GetInstance()->GetParameter(kSizeMaxParamId)->AddListener(this);
      ParameterManager::GetInstance()->GetParameter(kPitchMinParamId)->AddListener(this);
      ParameterManager::GetInstance()->GetParameter(kPitchMaxParamId)->AddListener(this);
      ParameterManager::GetInstance()->GetParameter(kDensityParamId)->AddListener(this);
      
      const int numGrains = 5;
      fDrone.SetNumGrains(numGrains);
      fDrone.SetIsTheDrone(true);
      fM.SetA(&fDrone);
      fM.SetVal(1.f/numGrains);
      AudioServer::GetInstance()->AddClient(&fM, 0);
      AudioServer::GetInstance()->AddClient(&fM, 1);
      AudioServer::GetInstance()->AddClient(this, 0);
      AudioServer::GetInstance()->AddClient(this, 1);
   }
   
   static Insaniac* GetInstance()
	{
		if (!sInstance)
		{
			sInstance = new Insaniac;
		}
		return sInstance;
	}

   // ParameterListener API
   virtual void Listen(Parameter const* p)
   {
      switch (p->Id())
      {
         case kSizeMinParamId:
            fDrone.SetMinLength(powf(p->Value(), 2) * 10);
            break;

         case kSizeMaxParamId:
            fDrone.SetMaxLength(powf(p->Value(), 2) * 10);
            break;
            
         case kPitchMinParamId:
            fDrone.SetMinFreq(p->Value() * 5 + 1);
            break;
            
         case kPitchMaxParamId:
            fDrone.SetMaxFreq(p->Value() * 5 + 1);
            break;
            
         case kDensityParamId:
            fDrone.SetDensity(p->Value());
            break;

         default:
            break;
      }
   }
   
   void AddCloudEvent(CloudEvent* e)
   {
      AudioServer::GetInstance()->EnterLock();
      fCloudEvents.push_back(e);
      AudioServer::GetInstance()->ExitLock();
   }
   
   void RemoveCloudEvent(CloudEvent* e)
   {
      AudioServer::GetInstance()->EnterLock();
      std::vector<CloudEvent*>::iterator i;
      i = std::find(fCloudEvents.begin(), fCloudEvents.end(), e);
      if (i != fCloudEvents.end())
      {
         delete *i;
         fCloudEvents.erase(i);
      }
      AudioServer::GetInstance()->ExitLock();
   }
   
   void AddNoiseBlip()
   {
      fNoiseBlips.push_back(new NoiseBlip);
   }
   
   // AudioClient API
   virtual void Render(float* buffer, int frames)
   {
      float tmp[frames];
      memset(tmp, 0.f, frames * sizeof(float));
      
      for (std::vector<CloudEvent*>::iterator i = fCloudEvents.begin(); i != fCloudEvents.end(); ++i)
      {
         (*i)->Render(tmp, frames);
         for (int i = 0; i < frames; ++i)
         {
            buffer[i] += tmp[i];
         }
      }
      
      for (std::vector<NoiseBlip*>::iterator i = fNoiseBlips.begin(); i != fNoiseBlips.end(); )
      {
         (*i)->Render(tmp, frames);
         for (int i = 0; i < frames; ++i)
         {
            buffer[i] += tmp[i];
         }
         
         if ((*i)->Done())
         {
            delete *i;
            i = fNoiseBlips.erase(i);
         }
         else
            i++;
      }
      
   }
   
private:
   Cloud fDrone;
   Multiplier fM;
   std::vector<CloudEvent*> fCloudEvents;
   std::vector<NoiseBlip*> fNoiseBlips;
   
   static Insaniac* sInstance;
};

#endif