#ifndef h_SignalGenerators
#define h_SignalGenerators

#include <cmath>
#include <cstdlib>
#include <queue>
#include <pthread.h>

#include "AudioClient.h"
#include "AudioServer.h"
#include "MathHelpers.h"

// Noise Source
// ----------------
// - White noise generation pilfered from http://musicdsp.org/showone.php?id=113
//
// - Pink noise not yet implemented
//
class NoiseSource : public AudioClient
{
public:
	NoiseSource(float gain = 1.f, int color = kWhite)
	: fGain(gain)
	, fColor(color)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		const float fs = AudioServer::GetInstance()->Fs();
		
		for (int i = 0; i < frames; ++i)
		{
			float R1 = (float) rand() / (float) RAND_MAX;
			float R2 = (float) rand() / (float) RAND_MAX;
			
			buffer[i] = tanh((float) sqrt( -2.0f * log( R1 )) * cos( 2.0f * PI * R2 ));
		}
	}
	
	enum NoiseColor
	{
		kWhite = 0,
		kPink,
		
		kNumNoiseColors
	};
	
private:
	float fGain;
	int fColor;
};

// Input Source
// ----------------
// - Provides one channel of input data from the Server
//
class InputSource : public AudioClient
{
public:
	InputSource(int channel = 0)
	: fChannel(channel)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		float inputBuffer[frames];
		AudioServer::GetInstance()->GetInput(inputBuffer, frames, fChannel);
		for (int i = 0; i < frames; ++i)
		{
			buffer[i] = inputBuffer[i];
		}
	}
	
private:
	int fChannel;
};

// SampleAccumulator
// --------------------
//  - Stores all samples that come in and provides access to them
//  - Stored buffer is cleared when accessed
//
class SampleAccumulator : public AudioClient
{
public:
	SampleAccumulator()
   : fSamplesPerPixel(12)
   , fMinSample(0.f)
   , fMaxSample(0.f)
   , fCount(0)
	{
      pthread_mutex_init(&fLock, NULL);
	}
   
   ~SampleAccumulator()
   {
      pthread_mutex_destroy(&fLock);
   }
	
	void Render(float* buffer, int frames)
	{
      
      if (fInput == NULL)
         return;

      fInput->Process(buffer, frames);

		for (int i = 0; i < frames; ++i)
		{
         if (fCount >= (fSamplesPerPixel - 1))
         {
            pthread_mutex_lock(&fLock);
            fPeakBuffer.push_back(std::make_pair(fMaxSample, fMinSample));
            pthread_mutex_unlock(&fLock);
            fMinSample = buffer[i];
            fMaxSample = buffer[i];
            fCount = 0;
         }
         else
         {
            if (buffer[i] > fMaxSample)
            {
               fMaxSample = buffer[i];
            }
            else if (buffer[i] < fMinSample)
            {
               fMinSample = buffer[i];
            }
            ++fCount;
         }
      }
   }
   
   typedef std::pair<float, float> PeakSample;
   typedef std::deque<PeakSample> PeakBuffer;
   
   PeakBuffer Get()
   {
      PeakBuffer samples;
      pthread_mutex_lock(&fLock);
      samples = fPeakBuffer;
      fPeakBuffer.clear();
      pthread_mutex_unlock(&fLock);
      return samples;
   }
   
   void SetInput(AudioClient* input)
   {
      fInput = input;
   }

   void SetSamplesPerPixel(int samplesPerPixel)
   {
      fSamplesPerPixel = samplesPerPixel;
   }

private:
   
   PeakBuffer fPeakBuffer;
   pthread_mutex_t fLock;
   int fSamplesPerPixel;
   float fMinSample;
   float fMaxSample;
   int fCount;

   AudioClient* fInput;
};

// Oscillator
// ----------------
// - Base class for oscillators.  Just a basic set of parameters to avoid
//   code duplication.
//
class Oscillator : public AudioClient
{
public:
	Oscillator(float freq = 440.f, float gain = 1.f, float width = 0.5f)
	: fFreq(freq)
	, fGain(gain)
	, fWidth(width)
   , fFreqZ(0)
   , fGainZ(0)
	, fT(0)
   , fPhase(0)
	{
	}
	
	void SetFreq(float freq)
	{
		fFreq = freq;
	}
	
	void SetGain(float g)
	{
		fGain = g;
	}
	
	void SetWidth(float w)
	{
		fWidth = w;
	}
	
protected:
	float fFreq;
	float fGain;
	float fWidth;
   
	float fT; // delete me
   float fPhase;
   
   float fFreqZ;
   float fGainZ;
};

// SinOsc
// ----------------
// - Generates a sine wave
//  
// - Would be nice to provide selectable approximations for efficiency
//
class SinOsc : public Oscillator
{
public:
	SinOsc(float freq = 440.f, float gain = 1.f)
	: Oscillator(freq, gain)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		const float fs = AudioServer::GetInstance()->Fs();
		const int period = 1.f / fFreqZ * fs;
		
		for (int i = 0; i < frames; ++i)
		{
			buffer[i] = fGainZ * sin(fPhase);
			
			fPhase += fFreqZ * (2*PI / fs);
			if (fPhase >= 2*PI)
				fPhase -= 2*PI;
         
         //fFreqZ = fFreqZ * 0.999 + fFreq * 0.001;
         //fGainZ = fGainZ * 0.999 + fGain * 0.001;
         fFreqZ = fFreq;
         fGainZ = fGain;
		}
	}
};

// FMOsc
// ----------------
// - Two-operator FM oscillator
//
class FMOsc : public Oscillator
{
public:
	FMOsc(float freq = 440.f, float gain = 1.f)
	: Oscillator(freq, gain)
	{
      fModOsc = new SinOsc(100.f);
      fModOsc->SetGain(0);
	}
	
   ~FMOsc()
   {
      delete fModOsc;
   }
   
	void Render(float* buffer, int frames)
	{
      
		const float fs = AudioServer::GetInstance()->Fs();
		const int period = 1.f / fFreqZ * fs;
		
      float modBuffer[frames];
      fModOsc->Render(modBuffer, frames);
      
		for (int i = 0; i < frames; ++i)
		{
			buffer[i] = fGainZ * sin(fPhase + modBuffer[i]);
			
         fPhase += fFreqZ * (2*PI / fs);
			if (fPhase >= 2*PI)
				fPhase -= 2*PI;
         
         fFreqZ = fFreqZ * 0.999 + fFreq * 0.001;
         fGainZ = fGainZ * 0.999 + fGain * 0.001;
		}
	}
   
   void SetModIndex(float index)
   {
      fModOsc->SetGain(index);
   }
   
   void SetModFreq(float freq)
   {
      fModOsc->SetFreq(freq);
   }
   
private:
   SinOsc *fModOsc;
};

// SawOsc
// ----------------
// - Generates a naive sawtooth (ramp)
//  
// - width parameter for changing the shape from ramp to triangle to inverse ramp
//
class SawOsc : public Oscillator
{
public:
	SawOsc(float freq = 440.f, float gain = 1.f, float width = 0.f)
	: Oscillator(freq, gain, width)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		const float fs = AudioServer::GetInstance()->Fs();
		
		int period = 1.f / fFreq * fs;
		int rising = period * (1 - fWidth);
		int falling = period - rising;
		
		for (int i = 0; i < frames; ++i)
		{
			if (fT < rising)
			{
				buffer[i] = 2 * fT / (float)rising - 1;
			}
			else
			{
				buffer[i] = 2 * (1 - ((fT - rising) / (float)falling)) - 1;
			}
			++fT;
			if (fT >= period)
				fT = 0;
		}
	}
};

// PwmOsc
// ----------------
// - Naive pulse width modulation oscillator class
//
class PwmOsc : public Oscillator
{
public:
	PwmOsc(float freq = 440.f, float gain = 1.f, float width = 5.f)
	: Oscillator(freq, gain, width)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		const float fs = AudioServer::GetInstance()->Fs();
		
		int period = 1.f / fFreq * fs;
		int pos = period * (1 - fWidth);
		
		for (int i = 0; i < frames; ++i)
		{
			if (fT < pos)
			{
				buffer[i] = 1.f;
			}
			else
			{
				buffer[i] = 0.f;
			}
			
			++fT;
			if (fT >= period)
				fT = 0;
		}
	}
};

// PulseTrain
// ----------------
// - Generates impulses at a given frequency
//
class PulseTrain : public Oscillator
{
public:
	PulseTrain(float freq = 440.f, float gain = 1.f, float width = 0.f)
	: Oscillator(freq, gain, width)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		const float fs = AudioServer::GetInstance()->Fs();
		
		int period = 1.f / fFreq * fs;
		
		for (int i = 0; i < frames; ++i)
		{
			if (fT == 0)
			{
				buffer[i] = 1.f;
			}
			
			++fT;
			if (fT >= period)
				fT = 0;
		}
	}
};

// AdditiveSinOsc
// ----------------
// - Uses an array of SinOscs to recreate the harmonic series for sawtooth and
//   square waves.
//
// - Selection of square vs. saw unimplemented
//
class AdditiveSinOsc : public Oscillator
{
public:
	AdditiveSinOsc(float freq = 440.f, float gain = 1.f, int order = 1)
	: Oscillator(freq, gain)
	, fOrder(order)
	{
		for (int i = 0; i < order; ++i)
		{
			// saw
			fSinOscs.push_back(new SinOsc(fFreq * (i+1), 1.f / (i+1)));
			
			// square
			//fSinOscs.push_back(new SinOsc(fFreq * (i * 2 + 1), 1.f / (i * 2 + 1)));
		}
	}
	
	~AdditiveSinOsc()
	{
		std::vector<SinOsc*>::iterator i;
		for (i  = fSinOscs.begin(); i != fSinOscs.end(); ++i)
		{
			delete (*i);
		}
		
		fSinOscs.clear();
	}
	
	void Render(float* buffer, int frames)
	{
		float tmp[frames];
		
		std::vector<SinOsc*>::iterator i;
		for (i = fSinOscs.begin(); i != fSinOscs.end(); ++i)
		{
			memset(tmp, 0.f, frames * sizeof(float));
			
			(*i)->Render(tmp, frames);
			
			for (int j = 0; j < frames; ++j)
			{
				buffer[j] += tmp[j];
			}
		}
	}
	
	void SetFreq(float freq)
	{
		std::vector<SinOsc*>::iterator i;
		int n = 0;
		for (i = fSinOscs.begin(); i != fSinOscs.end(); ++i)
		{
			(*i)->SetFreq(freq * (++n));
		}
		
		fFreq = freq;
	}
	
private:
	float fOrder;
	std::vector<SinOsc*> fSinOscs;
};


// Multiplier
// ----------------
// - Multiplier either multiplies two signals (fA * fB) or one signal by a constant value (fA * fConst)
//
// - the constant is used when fB == NULL
//
class Multiplier : public AudioClient
{
public:
	Multiplier(AudioClient* a = NULL, AudioClient* b = NULL, float constVal = 1.f)
	: fA(a)
	, fB(b)
	, fConst(constVal)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		if (fA)
		{
			fA->Process(buffer, frames);
			
			if (!fB)
			{
				for (int i = 0; i < frames; ++i)
				{
					buffer[i] *= fConst;
				}
			}
			else
			{
				float tmp[frames];
				fB->Process(tmp, frames);
				for (int i = 0; i < frames; ++i)
				{
					buffer[i] *= tmp[i];
				}
			}
		}
	}
	
	void SetA(AudioClient* a)
	{
		fA = a;
	}
	
	void SetB(AudioClient* b)
	{
		fB = b;
	}
	
	void SetVal(float val)
	{
		fConst = val;
	}
	
private:
	AudioClient* fA;
	AudioClient* fB;
	float fConst;
};

// Adder
// ----------------
// - Adder sums signals, and applies a constant offset (default 0) to the summed signals
//
// - Signals cannot be added to themselves
//
class Adder : public AudioClient
{
public:
	Adder(float constVal = 0.f)
	: fConst(constVal)
	{
	}
	
	void Render(float* buffer, int frames)
	{
		float tmp[frames];
		memset(tmp, 0.f, frames * sizeof(float));
		
		std::vector<AudioClient*>::iterator i;
		for (i = fClients.begin(); i != fClients.end(); ++i)
		{
			(*i)->Process(buffer, frames);
			for (int i = 0; i < frames; ++i)
			{
				tmp[i] += buffer[i];
			}
		}
		
		for (int i = 0; i < frames; ++i)
		{
			buffer[i] = tmp[i] + fConst;
		}
	}
	
	void SetVal(float val)
	{
		fConst = val;
	}
	
	void AddInput(AudioClient* c)
	{
		std::vector<AudioClient*>::iterator i = std::find(fClients.begin(), fClients.end(), c);
		if (i == fClients.end())
		{
			fClients.push_back(c);
		}
	}
	
	void RemoveInput(AudioClient* c)
	{
		std::vector<AudioClient*>::iterator i = std::find(fClients.begin(), fClients.end(), c);
		if (i != fClients.end())
		{
			fClients.erase(i);
		}
	}
	
private:
	std::vector<AudioClient*> fClients;
	float fConst;
};



#endif